(* **************************************************************************
 *
 * Elementary abstraction functions: single theorems, ops, types.
 *
 * $Source: /home/uniform/awe/repository/src/Extensions/abstract.sml,v $
 * $Revision: 784 $ of $Date: 2007-04-25 14:05:19 +0200 (Mi, 25 Apr 2007) $
 *)

signature ABSTRACTION =
sig
  val abstract_thm  : theory-> string-> thm-> thm
  val abstract_op   : theory-> string * typ-> string -> thm -> thm
  val abstract_type : theory-> typ-> string* sort-> thm-> thm
  val abstract_thy  : theory-> thm-> thm
  val abstract_map_thy  : theory-> (Proofterm.proof-> Proofterm.proof)-> 
                                   thm-> thm
end    

structure Abstract : ABSTRACTION  =
struct

(* Debugging*)
val occref = ref ([]: (Proofterm.proof* int) list)
val prfref = ref (#2 (#der (rep_thm trans)))

local open Aux in

(* ---------------------------------------------------------------------------
 *
 * Abstracting theorems
 *
 *)

(* For a list of variables (or occurences or something), calculate
 * deBruijn-indices, e.g. 
 * calc_index indexes ["a", "b", "c", "a", "x", "b"];
 * val it = [("a",0),("b",1),("c",2),("a",0),("x",3),("b",1)]
 *)

fun calc_indexes vars =
    let fun pair _ [] = []
	  | pair i ((c,cx)::rest) = 
	    if cx= (~1) then
		(c, i)::pair (i+1)
		           (* replace all other c with same index *)
		            (map (fn (d, dx)=> (d, if d= c then i else dx)) 
			     rest)
	    else (* This has already been replaced *)
		 (c, cx)::(pair i rest)
    in  pair 0 (vars ~~ (replicate (length vars) (~1)))
    end

(* Find all occurences of PThm(name,..) in proof term *)
(* NEW VERSION: Added case for PThm(("",_), ...)      *)

fun find_thm name (p as PThm(("", _), prf, _, _))= find_thm name prf
  | find_thm name (p as PThm((thmnm, _), _, _, _))= 
    if name= thmnm then [p] else []
  | find_thm name (Abst(_, _, p))= find_thm name p
  | find_thm name (AbsP(_, _, p))= find_thm name p
  | find_thm name (p % _)        = find_thm name p
  | find_thm name (p %% q)       = find_thm name p@ find_thm name q
  | find_thm _    _              = []


(* This function does one actual abstraction.
 * 
 * We replace one occurence of 
 *        PThm(name,...) % t1 % ... % tn %% p1 %% ... %% pm
 * with 
 *        PBound (index+count)% t1 % ... % tn %% p1 %% ... %% pm
 * so we count how many "t" and "p" we have to drop outside of the 
 * theorem (these are the drop_t and drop_p counters). abs is a flag 
 * saying wether we have abstracted anything yet.
 *)
fun abs_thm index name prf =
    let	fun const p = {prf= p, drop_t= 0, drop_p= 0, abs= false}
        fun abs_thm'  c ts ps (prf as PThm((thmnm, tags), p, t, ty)) =
	     if  name= thmnm then 
		 {prf= foldl (op %%) (foldl (op %) (PBound c, ts), ps),
		  drop_t = length ts,
		  drop_p = length ps,
		  abs= true}
	     else const prf
	  | abs_thm'  c _ _   (p as PBound _)   = const p
	  | abs_thm'  c ts ps (Abst(nm, ty, p)) = 
	    let val {prf, drop_t, drop_p, abs}=  abs_thm'  c ts ps p
	    in  {prf= Abst(nm, ty, prf), 
		 drop_t= drop_t, drop_p= drop_p, abs= abs}
	    end
	  | abs_thm'  c ts ps (AbsP(nm, t, p))  = 
	    let val {prf, drop_t, drop_p, abs}= abs_thm' (c+1) ts ps p
	    in  {prf= AbsP(nm,  t, prf),
		 drop_t= drop_t, drop_p= drop_p, abs= abs}
	    end
	  | abs_thm'  c ts ps (p % t)  = 
	    let val {prf,drop_t, drop_p, abs}= abs_thm' c (t::ts) ps p
	    in  {prf =   if drop_t > 0 then prf else prf % t,
		 drop_t= if drop_t > 0 then drop_t-1 else 0,
		 drop_p= drop_p, abs= abs}
	    end		 
	  | abs_thm'  c ts ps (p %% q) = 
	    let val {prf, drop_t, drop_p, abs}= abs_thm' c [] (q::ps) p
	    in  if abs then 
		{prf= if drop_p > 0 then prf else prf %% q,
	         drop_t= 0,
		 drop_p= if drop_p > 0 then drop_p-1 else 0,
		 abs= true}
		else
		    let val {prf= qprf, drop_t=dt, drop_p=dp, abs}= 
			abs_thm' c [] [] q
		    in  {prf=p %% qprf, drop_t= dt, drop_p= dp, abs= abs}
		    end
	    end
	  | abs_thm'  c ts ps (h as Hyp _)     = const h
	  | abs_thm'  c ts ps (ax as PAxm _)   = const ax
	  | abs_thm'  c ts ps (or as Oracle _) = const or
	  | abs_thm'  c ts ps (p as MinProof _)= const p
    in
	#prf (abs_thm' index [] [] prf)
    end

(* Main functions:
 *  abstract one or many thms, with or without explicit theory argument
 *)
fun abstract_thm thy name thm =
    let val {der=(_, prf), prop= t, sign=sg, ...} = rep_thm thm

	val absthm = Aux.get_thm thy name 

	val thmprf = Aux.strip_thm prf

	val fullprf = (Aux.fullprf_of thm) 

	(* find all occurences, calculate indices *)
        val ocs = calc_indexes (find_thm name fullprf)
	val _ = occref := ocs;

	(* do the proof abstraction-- this only needs the indices *)
        val absprf  = foldl (fn (prf, ix)=> abs_thm ix name prf)
	                    (thmprf, map snd ocs)

        (* Construct one premise-- need to instantiate type variables! *)
	fun add_prems ((thmprf, t), (pf as PThm(_, _, _, ts))) = 
	    let val instthm = 
		 case ts of
                     SOME ts => 
			 let (* val typinst = map fst (term_tvars 
						    (prop_of absthm)) ~~
			                   map (Thm.ctyp_of sg) ts *)

                             val typinst = (map ((Thm.ctyp_of sg) o TVar)
                                                (term_tvars 
						    (prop_of absthm))) ~~
			                   (map (Thm.ctyp_of sg) ts)

			 in  Thm.instantiate (typinst, []) 
			           (forall_intr_vars absthm)
			 end
		   | NONE => absthm
  		val nuprem  = prop_of instthm (* (forall_intr_vars instthm) *)
		val absprf  = AbsP("H", NONE, thmprf)
		val absprop = Const("==>", propT--> propT--> propT) $ nuprem $ t
	    in  (absprf, absprop)
	    end
	(* calculate the new proposition, and the AbsP's; 
	 * this only needs the typ instances, and we have to get rid of
	 * multiple occurences; the indices have been calculated in such
	 * a way that only the first occurence is relevant. *)
	val (absprf, absprop) = 
	               foldl add_prems ((absprf, t), distinct (map fst ocs))
	val _  = prfref := absprf
	val fullabsprf = Aux.reconstruct_prf sg absprop absprf 

    in  Derive_Thm.derive_thmI thy fullabsprf
(* Aux.derive_thm thy fullabsprf *)
    end

(* --------------------------------------------------------------------------
 *
 * Abstract over operations 
 * 
 *)

fun abstract_op thy (name, typ) varname thm =
    let fun abs_term (s $ t) = abs_term s $ abs_term t
          | abs_term (Abs(s, ty, t)) = Abs(s, ty, abs_term t)
          | abs_term (c as Const(nm, ty)) = 
                     if nm = name andalso ty= typ then Free(varname, ty) else c
          | abs_term t               = t (* Free, Var, Bound *)

	val fullprf    = (Aux.fullprf_of thm)
	val absprf     = Proofterm.map_proof_terms abs_term I fullprf
    in  Aux.derive_thm thy absprf
    end


(* --------------------------------------------------------------------------
 *
 * Abstract over types
 *
 *)

fun repl_typ (old, nu) t =
    (if t= old then nu else
	 case t of 
	     Type(nm, ts)    => Type(nm, map (repl_typ (old, nu)) ts)
	   | (tf as TFree _) => tf
	   | (tv as TVar _)  => tv)
               
fun abstract_type nuthy old (nunm, nusorts) thm =
    let val nut = TFree(nunm, nusorts)
        val fullprf  = (Aux.fullprf_of thm)
        val absprf   = Proofterm.map_proof_terms
			   (map_term_types (repl_typ (old, nut)))
	                   (repl_typ (old, nut)) fullprf
    in  Aux.derive_thm nuthy absprf
    end



(* --------------------------------------------------------------------------
 *
 * Abstract over theories
 *
 *)

fun abstract_map_thy nuthy map thm = 
    Aux.derive_thm nuthy (map (Aux.fullprf_of thm))

fun abstract_thy nuthy = abstract_map_thy nuthy I


end (* local open Aux *)
end 

