

signature THYMORPH =
sig
 type thymorph
 datatype ThyMorphData = ThyMorphData of 
                          { dom_thms : (Thm.thm * Term.term * Symtab.key list) Symtab.table,
                            cod_thms : (Thm.thm * Term.term) Symtab.table }

 val data : thymorph -> ThyMorphData
 val get_thymorph_data : Sigmorph.sigmorph -> ThyMorphData
 val sigmorph : thymorph -> Sigmorph.sigmorph
 val axiomMap : thymorph ->  Derive_Thm.AxmImg Symtab.table
 val theoremMap : thymorph -> Symtab.key Symtab.table
 val match_thms : thymorph -> thymorph
 val make : theory -> theory -> (string * string) list -> (string * string) list ->
            (string * string) list -> thymorph
 val make_by_sigmorph : Sigmorph.sigmorph -> (string * string) list -> thymorph
 val normalize : (string * mixfix option) Symtab.table -> thymorph -> thymorph
 val translate_syntax : (string * mixfix option) Symtab.table -> thymorph -> thymorph
 val translate_thm : thymorph -> string -> string -> (thymorph * ((string * Thm.thm) list))
 val compose : thymorph -> thymorph -> thymorph
 val cod_ext : thymorph -> Symtab.key Symtab.table -> theory -> thymorph


 datatype Inst_Result = ThyM of thymorph 
                      | SigM of (Sigmorph.sigmorph * ((string * Thm.cterm) list))

 val instantiate : bool -> bool -> (string * string) list -> (string * string) list -> (string * string) list ->
                  (string * mixfix option) Symtab.table -> 
                   theory -> thymorph -> Inst_Result

 val thymorph_goals : bool -> Sigmorph.sigmorph -> (Sigmorph.sigmorph * ((string * Thm.cterm) list)) 

 val add_thymorph : (string * thymorph) -> unit
 val get_thymorph : string -> thymorph option
 val print_thymorphs : unit -> string
 
 exception THYMORPH of string

end


structure Thymorph : THYMORPH =
struct

 exception THYMORPH of string

 datatype ThyMorphData = ThyMorphData of 
                          { dom_thms : (Thm.thm * Term.term * Symtab.key list) Symtab.table,
                            cod_thms : (Thm.thm * Term.term) Symtab.table }




 datatype thymorph = ThyMorph of { sigma : Sigmorph.sigmorph,
                                   axiomMap : Derive_Thm.AxmImg Symtab.table,
                                   theoremMap : Symtab.key Symtab.table,
                                   data : ThyMorphData }



fun sigmorph (ThyMorph{ sigma, ... }) = sigma
fun axiomMap (ThyMorph{ axiomMap, ... }) = axiomMap
fun theoremMap (ThyMorph{ theoremMap, ...}) = theoremMap
fun data (ThyMorph{ data, ... }) = data



     fun un_str (a : string list, b) = a union b
     fun diff_str (s : string list) t = s \\ t

     fun assoc_defs defnm_rel dep_ops = Aux.foldl un_str ([], map (Symtab.lookup_multi defnm_rel) dep_ops)     


     fun ext_mapping mp [] = mp
       | ext_mapping mp ((x as (key, v))::xs) =
         let val mp' = Symtab.update_new x mp
                       handle DUP => (let val _ = writeln ("Updating " ^ key)
                                      in
                                       Symtab.update x mp
                                      end)
         in
          ext_mapping mp' xs
         end

   
     fun prepare_dom_thms dg t [] = t
       | prepare_dom_thms dg t ((nm, thm)::xs) =
         let val norm = Aux.unvarify o prop_of o Aux.beta_eta_convert
             val thm_dep = (case (Symtab.lookup dg nm) of
                             NONE => []
                           | SOME (_, deps) => deps)
         in
          prepare_dom_thms dg (Symtab.update (nm, (thm, norm thm, thm_dep)) t) xs
         end 
     

     fun prepare_cod_thms t [] = t
       | prepare_cod_thms t ((nm, thm)::xs) =
         let val norm = Aux.unvarify o prop_of o Aux.beta_eta_convert
         in
          prepare_cod_thms (Symtab.update (nm, (thm, norm thm)) t)xs
         end      




fun get_thymorph_data sigma =
   let val Sigmorph.SigMorphData d = Sigmorph.data sigma
       val src = Sigmorph.src sigma
       val trg = Sigmorph.trg sigma     
       val src_id = Aux.name_of_thy src
       val trg_id = Aux.name_of_thy trg
       val dom = #dom d
       val cod = #cod d
       val domain_thys = (if (dom = []) then [] else src::(map theory (dom \\ [src_id])))
       val codomain_thys = (if (cod = []) then [] else trg::(map theory (cod \\ [trg_id])))
       val dom_thms = (Aux.foldl (op @) ([], map Aux.thms_of domain_thys))
       val cod_thms = (Aux.foldl (op @) ([], map Aux.thms_of codomain_thys))

       val dom_thms_dep = Dependency.thms_dep_graph dom_thms
       val dom_thms' = prepare_dom_thms dom_thms_dep Symtab.empty dom_thms 
       val cod_thms' = prepare_cod_thms Symtab.empty cod_thms
   in
    ThyMorphData { dom_thms = dom_thms', cod_thms = cod_thms' }
   end




(* **************************** *)
(* *** Theorem matching ******* *)
(* **************************** *)

fun match_thms_i sigma thmMap dom_thms cod_thms = 
  let val trg = Sigmorph.trg sigma
      val tsig = Sign.tsig_of trg

      fun match_thm tsig (nm, term) [] = NONE
        | match_thm tsig (nm, term) ((x as (nm', (thm', term')))::xs) =
            if (Alpha.term_aconv tsig (term, term')) 
            then (let val _ = writeln ("Found: (" ^ (quote nm) ^ " |-> " ^ (quote nm') ^ ")")
                  in
                   SOME (nm, nm')
                  end)
            else match_thm tsig (nm, term) xs
 
      fun match_thms_ tsig ext cod_thms [] = ext
        | match_thms_ tsig ext cod_thms ((nm, (thm, term, _))::xs) =
          (case (Symtab.lookup ext nm) of
            NONE => (let (* val _ = writeln ("Checking thm: " ^ nm) *)
                         val term' = Sigmorph.map_term sigma term
                     in
                       case (match_thm tsig (nm, term') cod_thms) of
                        NONE => match_thms_ tsig ext cod_thms xs
                      | SOME mp => (let val ext' = ext_mapping ext [mp]
                                    in
                                     match_thms_ tsig ext' cod_thms xs
                                    end)
                     end)
          | _ => match_thms_ tsig ext cod_thms xs)
  in
   match_thms_ tsig thmMap cod_thms dom_thms
  end   

fun match_thms (ThyMorph{sigma, axiomMap, theoremMap, data = ThyMorphData d}) =
   let val dom_thms = Symtab.dest (#dom_thms d)
       val cod_thms = Symtab.dest (#cod_thms d)      
       val theoremMap' = match_thms_i sigma theoremMap dom_thms cod_thms
   in
     ThyMorph{sigma = sigma, axiomMap = axiomMap, 
              theoremMap = theoremMap', data = ThyMorphData d}
   end



(* **************************** *)
(* *** Axiom matching ********* *)
(* **************************** *)

fun match_axms all sigma axiomMap cod_thms =
 let val Sigmorph.SigMorphData d = Sigmorph.data sigma
     val dom_axms = Symtab.dest (#dom_axs_map d)
     val cod_axms = Symtab.dest (#cod_axs_map d)
     val trg = (Sigmorph.trg sigma)
     val tsig = Sign.tsig_of trg

     fun match_by_ax term [] = NONE
       | match_by_ax term ((nm, term')::xs) =
            (if (Alpha.term_aconv tsig (term, term')) then SOME nm 
             else match_by_ax term xs)
     fun match_by_thm term [] = NONE
       | match_by_thm term ((nm, (thm, term'))::xs) =
            (if (Alpha.term_aconv tsig (term, term')) then SOME nm 
             else match_by_thm term xs)
 

     fun match_ax axMap notmapped [] = (axMap, notmapped)
       | match_ax axMap notmapped ((nm, term)::xs) =
       (case (Symtab.lookup axMap nm) of
         NONE => (let val term' = Sigmorph.map_term sigma term 
                  in
                   case (match_by_ax term' cod_axms) of
                     NONE => (case (match_by_thm term' cod_thms) of
                               NONE => match_ax axMap (nm::notmapped) xs
                             | SOME i => (let val _ = Aux.info ("Axiom mapping found: (" ^ 
                                                                 (quote nm) ^ " |-> " ^ (quote i) ^ ")")
                                          in
                                           match_ax (Symtab.update_new (nm, Derive_Thm.Thm i) axMap) notmapped xs
                                          end))
                   | SOME i => (let val _ = Aux.info ("Axiom mapping found: (" ^ 
                                                      (quote nm) ^ " |-> " ^ (quote i) ^ ")")
                                in
                                 match_ax (Symtab.update_new (nm, Derive_Thm.Axm i) axMap) notmapped xs
                                end)
                  end)
      | SOME i => match_ax axMap notmapped xs)

     val (axMap, notmapped) = match_ax axiomMap [] dom_axms 
 in
  if (not all) then axMap
  else 
   (case notmapped of
     [] => axMap
    | _ =>  (let val _ = Aux.error_msg ("Cannot match following axioms: " ^ (commas notmapped))
             in
              raise THYMORPH "Theory morphism construction failed."
             end))
 end 

(* **************************** *)
(* *** Representation   ******* *)
(* **************************** *)

   fun show_map_i [] = ""
     | show_map_i [(s, t)] = ("(" ^ (quote s) ^ " |-> " ^ (quote t) ^ ")")
     | show_map_i ((s, t)::m) = (("(" ^ (quote s) ^ " |-> " ^ (quote t) ^ "),\n") 
                                  ^ (show_map_i m))


(* **************************** *)
(* *** Construction     ******* *)
(* **************************** *)

fun transf_axMap [] = []
  | transf_axMap ((a, b)::xs) = (case b of
                                    Derive_Thm.Axm b' => (a, b')::(transf_axMap xs)
                                  | Derive_Thm.Thm b' => (a, b')::(transf_axMap xs))


fun make_by_sigmorph sigma axMap_rep =
 let val Sigmorph.SigMorphData d = Sigmorph.data sigma
     val (td as (ThyMorphData { dom_thms, cod_thms })) = get_thymorph_data sigma
     val dom_axms = #dom_axs_map d
     val cod_axms = #cod_axs_map d
     val trg = Sigmorph.trg sigma
     val tsig = Sign.tsig_of trg

     fun mk_axiomMap err t [] = (err, t)
       | mk_axiomMap err t ((a, b)::xs) =
     (case (Symtab.lookup dom_axms a) of
      NONE => mk_axiomMap err t xs
    | SOME term => 
       (case (Symtab.lookup cod_axms b) of
         SOME term' => 
              (if (Alpha.term_aconv tsig (Sigmorph.map_term sigma term, term'))
               then mk_axiomMap err (Symtab.update (a, Derive_Thm.Axm b) t) xs
               else mk_axiomMap ((a, b)::err) t xs)
     | NONE => (case (Symtab.lookup cod_thms b) of
                 SOME (thm', term') => (if (Alpha.term_aconv tsig (Sigmorph.map_term sigma term, term'))
                                        then mk_axiomMap err (Symtab.update (a, Derive_Thm.Thm b) t) xs
                                        else mk_axiomMap ((a, b)::err) t xs) 
               | NONE => (let val pthm = (true, get_axiom trg b)
                                          handle e => (false, Aux.get_thm trg b)
                              val norm = Aux.unvarify o prop_of o Aux.beta_eta_convert
                          in
                           case pthm of
                            (true, thm') => (let val term' = norm thm'
                                             in
                                              (if (Alpha.term_aconv tsig (Sigmorph.map_term sigma term, term'))
                                               then mk_axiomMap err (Symtab.update (a, Derive_Thm.Axm b) t) xs
                                               else mk_axiomMap ((a, b)::err) t xs)
                                             end)
                         | (false, thm') => (let val term' = norm thm'
                                             in
                                              (if (Alpha.term_aconv tsig (Sigmorph.map_term sigma term, term'))
                                               then mk_axiomMap err (Symtab.update (a, Derive_Thm.Thm b) t) xs
                                               else mk_axiomMap ((a, b)::err) t xs)
                                             end)
                          end))))

     val axiomMap = (case (mk_axiomMap [] Symtab.empty axMap_rep) of
                      ([], t) => match_axms true sigma t (Symtab.dest cod_thms)
                    | (err, t) => 
                    (let val _ = Aux.error_msg ("Following axiom mapping(s) are illegal:\n" ^ 
                                                    show_map_i err)
                     in
                       raise THYMORPH "Theory morphism construction failed."
                     end))  
     val tmorph =   ThyMorph { sigma = sigma, axiomMap = axiomMap, 
                               theoremMap = Symtab.empty, data = td }
     val tmorph' = match_thms tmorph  
 in
  tmorph'
 end


fun make src trg tpMap_rep opMap_rep axMap_rep =
 let val sigma = Sigmorph.make src trg tpMap_rep opMap_rep
 in
  make_by_sigmorph sigma axMap_rep
 end


(* **************************** *)
(* *** Extension        ******* *)
(* **************************** *)


fun cod_ext (ThyMorph {sigma, axiomMap, theoremMap, data=ThyMorphData td}) 
            thmMap_ext ntrg =
 let val o_trg = Sigmorph.trg sigma
     val nsigma = Sigmorph.cod_ext sigma ntrg
     val Sigmorph.SigMorphData d = Sigmorph.data nsigma
     val n1 = Aux.name_of_thy o_trg
     val n2 = Aux.name_of_thy ntrg
 in 
   if (n1 = n2) then 
    (let val new_thms = (let val thms1 = map fst (Symtab.dest (#cod_thms td)) 
                             val thms2 = (Aux.thms_of ntrg)
                         in
                          filter (fn (nm, t) => (not (nm mem_string thms1))) thms2
                         end) 
         val new_thms' = prepare_cod_thms Symtab.empty new_thms
         val ndata = ThyMorphData {dom_thms = (#dom_thms td),
                                   cod_thms = Symtab.extend (#cod_thms td, Symtab.dest new_thms') }
                     handle DUPS => error ("DUPS1 in Thymorph.cod_ext")
         val tmorph = ThyMorph {sigma = nsigma, axiomMap = axiomMap, 
                                theoremMap = (Symtab.merge (op =) (theoremMap, thmMap_ext)), 
                                data = ndata}
                      handle DUPS => error ("DUPS2 in Thymorph.cod_ext")                      
         val tmorph' = (case new_thms of
                         [] => tmorph
                       | _ => match_thms tmorph)
     in
      tmorph'
     end)
   else (case (Aux.unq_path o_trg ntrg) of
          [] => make_by_sigmorph nsigma (transf_axMap (Symtab.dest axiomMap))
        | path => 
          if ((Aux.name_of_thy (Sigmorph.src nsigma)) mem_string (map Aux.name_of_thy path))
          then make_by_sigmorph nsigma (transf_axMap (Symtab.dest axiomMap))  
          else    (let val new_thms = Aux.foldl op@ ([], map (Aux.thms_of) path)  
                       val new_thms' =  prepare_cod_thms Symtab.empty new_thms
                       val ndata = ThyMorphData {dom_thms = (#dom_thms td),
                                                 cod_thms = Symtab.extend (#cod_thms td, Symtab.dest new_thms') }
                                   handle DUPS => error ("DUPS3 in Thymorph.cod_ext")
                       val tmorph = ThyMorph {sigma = nsigma, axiomMap = axiomMap, 
                                              theoremMap = (Symtab.merge (op =) (theoremMap, thmMap_ext)), 
                                              data = ndata}
                                    handle DUPS => error ("DUPS4 in Thymorph.cod_ext")                                  
                       val tmorph' = (case new_thms of
                                       [] => tmorph
                                     | _ => match_thms tmorph)
                   in
                    tmorph' 
                   end))
 end


(* ********************* *)
(* *** Normalization *** *)
(* ********************* *)

fun normalize renames (ThyMorph {sigma, axiomMap, theoremMap, data}) =
 let val nsigma = Sigmorph.normalize renames sigma
 in
  ThyMorph{sigma = nsigma, axiomMap=axiomMap, 
           theoremMap=theoremMap, data = data}
 end

fun translate_syntax renames (ThyMorph {sigma, axiomMap, theoremMap, data}) =
 let val nsigma = Sigmorph.translate_syntax renames sigma
 in
  ThyMorph{sigma = nsigma, axiomMap=axiomMap, 
           theoremMap=theoremMap, data = data}
 end



(* **************************** *)
(* *** Theorem translations *** *)
(* **************************** *)



fun new_name names pre id = (if (id mem_string names) then new_name names pre (pre ^ "_" ^ id)
                             else id)


fun translate_thms sat src thy trg_id trg_names pre dom_thms sigma axiomMap theoremMap [] = (thy, theoremMap)
  | translate_thms sat src thy trg_id trg_names pre dom_thms sigma axiomMap theoremMap (x::xs) =
(case (Symtab.lookup theoremMap x) of
   SOME i => translate_thms sat src thy trg_id trg_names pre dom_thms sigma axiomMap theoremMap xs
 | NONE => (case (Symtab.lookup dom_thms x) of
             NONE => translate_thms sat src thy trg_id trg_names pre dom_thms sigma axiomMap theoremMap xs
           | SOME (thm, term, dep) =>
         (if (x mem_string sat) then 
             (let val newname_ = new_name trg_names pre (snd (Aux.qname_split2 x))
                  val (path, newname) = Aux.qname_split newname_
                  val newname' = if path = "" then (trg_id ^ "." ^ newname)
                                 else (trg_id ^ "." ^ path ^ "." ^ newname)

                  val _ = Aux.info ("... translating theorem "^ x ^" as " ^ newname' ^ " ...") 

                  val newthm = (if !AWE.skip_proofs 
                                then (let val t = Aux.unvarify (prop_of thm)
                                          val mt = Sigmorph.map_term sigma t
                                          val _ = Aux.warning "proof skipped"
                                          val _ = (quick_and_dirty := true)
                                          val thm = (Drule.standard o Thm.varifyT) (SkipProof.make_thm thy mt) 
                                          val _ = (quick_and_dirty := false)
                                      in
                                       thm
                                     end)
                                else (let val prf = Aux.fullprf_of thm
                                          val defMap = Sigmorph.defMap sigma 
                                      in 
                                       Derive_Thm.derive_thm (Sigmorph.map_type sigma) (Sigmorph.map_term sigma)
                                                             defMap axiomMap theoremMap src thy prf
                                      end))

                  val theoremMap' = Symtab.update_new (x, newname') theoremMap
                                     handle DUP => error "Unexpected DUP in translate_thm_i"
                  val (thy', gen) = 
                  (if (path = "") then PureThy.store_thm (Thm.no_attributes(newname, newthm)) thy
                   else thy 
                        |> Sign.add_path path
                        |> PureThy.store_thm (Thm.no_attributes(newname, newthm))
                        |>> Sign.parent_path)
                  
                  val _ = Aux.info ("... added theorem " ^ newname' ^ " : \n" ^ (string_of_thm newthm))
              in
               translate_thms sat src thy' trg_id trg_names pre dom_thms sigma axiomMap theoremMap' xs
              end)
          else translate_thms (x::sat) src thy trg_id 
               trg_names pre dom_thms sigma axiomMap theoremMap (dep @ (x::xs)))))  



fun translate_thm_i (tmorph as (ThyMorph {sigma, axiomMap, theoremMap, data = ThyMorphData td, ...})) 
                     rename id =
(case (Symtab.lookup theoremMap id) of
  SOME i => (tmorph, SOME (i, Aux.get_thm (Sigmorph.trg sigma) i))
| NONE => (case (Symtab.lookup (#dom_thms td) id) of
             NONE => (let val thm = SOME (Aux.get_thm (Sigmorph.trg sigma) id)
                                    handle e => (let val _ = writeln ("Identifier " ^ id ^ 
                                                                      " is neither local nor global")
                                                 in
                                                  NONE
                                                 end)
                      in
                        case thm of
                          SOME t => (tmorph, SOME(id, t))
                        | NONE => (tmorph, NONE)
                      end)
           | SOME (thm, term, dep) => 
              (let val trg = Sigmorph.trg sigma
                   val trg_id = Aux.name_of_thy trg
                   val trg_names = map (snd o Aux.qname_split2 o fst) (Aux.thms_of trg)
                   val src = Sigmorph.src sigma
                   val src_id = Aux.name_of_thy src
                   val Sigmorph.SigMorphData d = Sigmorph.data sigma

                   val newname_ = (if rename = "" then new_name trg_names src_id (snd (Aux.qname_split2 id))
                                   else rename)
                   val (path, newname) = Aux.qname_split newname_
                   val newname' = if path = "" then (trg_id ^ "." ^ newname)
                                  else (trg_id ^ "." ^ path ^ "." ^ newname)

                   val _ = Aux.info ("Translating theorem "^ id ^" as " ^ newname' ^ " ...") 

                   val (trg', theoremMap') = translate_thms [] src trg trg_id trg_names src_id 
                                             (#dom_thms td) sigma axiomMap theoremMap dep
                   val newthm = (if !AWE.skip_proofs 
                                 then (let val t = Aux.unvarify (prop_of thm)
                                           val mt = Sigmorph.map_term sigma t
                                           val _ = Aux.warning "proof skipped"
                                           val _ = (quick_and_dirty := true)
                                           val thm = (Drule.standard o Thm.varifyT) (SkipProof.make_thm trg' mt) 
                                           val _ = (quick_and_dirty := false)
                                       in
                                        thm
                                      end)
                                 else (let val prf = Aux.fullprf_of thm  
                                           val defMap = Sigmorph.defMap sigma 
                                       in 
                                        Derive_Thm.derive_thm (Sigmorph.map_type sigma)(Sigmorph.map_term sigma) 
                                                             defMap axiomMap theoremMap' src trg' prf
                                       end))

                   val (trg'', gen) = 
                  (if (path = "") then PureThy.store_thm ((newname, newthm), []) trg'
                   else trg' 
                        |> Sign.add_path path
                        |> PureThy.store_thm ((newname, newthm), [])
                        |>> Sign.parent_path)                  
                   val theoremMap'' = Symtab.update_new (id, newname') theoremMap'
                                      handle DUP => error "Unexpected DUP in translate_thm_i"
                   val _ = Aux.info ("... added theorem " ^ newname' ^ " :\n" ^ (string_of_thm newthm))
                   val tmorph' = cod_ext tmorph theoremMap'' trg''
               in
                (tmorph', SOME (newname', newthm))
               end)))


fun translate_thm tmorph rename id =
 let val tmorph' = normalize Symtab.empty tmorph
     val sigma = sigmorph tmorph'
     val src = Sigmorph.src sigma
     val thms = PureThy.get_thms src (Name id)
 in
  (case thms of
     [] => error "no thms"
   | [x] => (let val (tmorph'', opt) = translate_thm_i tmorph' rename id
             in
              case opt of
               NONE => (tmorph'', [])
             | SOME r => (tmorph'', [r])
             end)
   
   | xs => 
     (let val ids = map Thm.name_of_thm xs
          fun translate ((mrph, translated), x) = 
             (let val (mrph', opt) = translate_thm_i mrph "" x
              in
               case opt of
                NONE => error "Unexpected error: empty result in translate_thm"
              | SOME r => (mrph', translated@[r])
              end)       
          val (tmorph'', tthms) = Aux.foldl translate ((tmorph', []), ids) 

          val trg = Sigmorph.trg (sigmorph tmorph'')
          val trg_id = Aux.name_of_thy trg
          val trg_names = map (snd o Aux.qname_split2 o fst) (Aux.thms_of trg)
          val src = Sigmorph.src (sigmorph tmorph'')
          val src_id = Aux.name_of_thy src

          val newname_ = (if rename = "" then new_name trg_names src_id (snd (Aux.qname_split2 id))
                          else rename)
          val (path, newname) = Aux.qname_split newname_
          val newname' = if path = "" then (trg_id ^ "." ^ newname)
                         else (trg_id ^ "." ^ path ^ "." ^ newname)
          val (trg', gen) = 
          (if (path = "") then PureThy.add_thmss [((newname, map snd tthms), [])] trg
           else trg
                |> Sign.add_path path
                |> PureThy.add_thmss [((newname, map snd tthms), [])] 
                |>> Sign.parent_path)

          val tmorph''' = cod_ext tmorph'' Symtab.empty trg'
          val _ = Aux.info ("Added lemmas " ^ newname')
     in
       (tmorph''', tthms)
     end))
 end

(* **************************** *)
(* *** Composition      ******* *)
(* **************************** *)


fun compose (tm1 as (ThyMorph {sigma=sigma1, axiomMap=axiomMap1, theoremMap=theoremMap1, ...}))
            (tm2 as (ThyMorph {sigma=sigma2, axiomMap=axiomMap2, ...})) =
 let val sc = Sigmorph.compose sigma1 sigma2
     val Sigmorph.SigMorphData d = Sigmorph.data sc
     val basis = Symtab.dest(#dom_axs_map d)

     fun mk_axMaps tm2u a [] = (tm2u, a)
       | mk_axMaps tm2u a ((nm, _)::bs) = 
     (case (Symtab.lookup axiomMap1 nm) of
      SOME (Derive_Thm.Axm i) => 
                 (case (Symtab.lookup axiomMap2 i) of
                         SOME (Derive_Thm.Axm i') => mk_axMaps tm2u 
                                                     (Symtab.update (nm, Derive_Thm.Axm i') a) bs
                       | SOME (Derive_Thm.Thm i') => mk_axMaps tm2u 
                                                     (Symtab.update (nm, Derive_Thm.Thm i') a) bs
                       | NONE =>  mk_axMaps tm2u (Symtab.update (nm, Derive_Thm.Axm i) a) bs)
    | SOME (Derive_Thm.Thm i) => 
                 (case (Symtab.lookup (theoremMap tm2u) i) of
                        SOME i' => mk_axMaps tm2u (Symtab.update (nm, Derive_Thm.Thm i') a) bs
                      | NONE => (let val (tm2u', i') = (case (translate_thm tm2u "" i) of
                                                         (t, [(x, _)]) => (t, x)
                                                       | (t, _) => error ("Unexpected result of theorem\
                                                                              \ translation in compose")) 
                                 in
                                  mk_axMaps tm2u' (Symtab.update (nm, Derive_Thm.Thm i') a) bs
                                 end))
    | NONE => (case (Symtab.lookup axiomMap2 nm) of
                SOME (Derive_Thm.Axm i') => mk_axMaps tm2u (Symtab.update (nm, Derive_Thm.Axm i') a) bs
              | SOME (Derive_Thm.Thm i') => mk_axMaps tm2u (Symtab.update (nm, Derive_Thm.Thm i') a) bs
              | NONE => error "Unexpected global axiom in compose: mk_axMaps"))

     val (tm2u, a) = mk_axMaps tm2 Symtab.empty basis
     val ntrg = Sigmorph.trg(sigmorph tm2u)
     val sc' = Sigmorph.cod_ext sc ntrg
     val td = get_thymorph_data sc'
     val theoremMap2 = theoremMap tm2u

     fun mk_thmMap m [] = m
       | mk_thmMap m ((a, b)::xs) = 
      (case (Symtab.lookup theoremMap2 b) of
        SOME i => (let val _ = Aux.info ("Composed: (" ^ (quote a)
                                            ^ " |-> " ^ (quote i) ^ ")")
                   in
                    mk_thmMap (Symtab.update (a, i) m) xs
                   end)
      | NONE => mk_thmMap m xs)

      val thmMap = mk_thmMap Symtab.empty (Symtab.dest theoremMap1)
      val ctmorph = ThyMorph{sigma=sc', axiomMap=a, theoremMap=thmMap, data = td}
 in
  match_thms ctmorph
 end
 



(* **************************** *)
(* *** Instantiation    ******* *)
(* **************************** *)

datatype Inst_Result = ThyM of thymorph 
                     | SigM of (Sigmorph.sigmorph * ((string * Thm.cterm) list))


fun instantiate axiomatic syn type_specs op_specs axm_specs renames thy 
     (tmorph as (ThyMorph {sigma, axiomMap=axiomMap_, theoremMap, ...})) =
let val (nsigma, axiomMap') = Sigmorph.instantiate syn type_specs op_specs renames thy sigma
    val Sigmorph.SigMorphData d = Sigmorph.data nsigma
    val (ThyMorphData td) = get_thymorph_data nsigma
    val axs_basis = Symtab.dest(#dom_axs_map d)
    val cod_thms = Symtab.dest(#cod_thms td)

    val axiomMap'' = ext_mapping axiomMap_ (Symtab.dest axiomMap')
    val axiomMap = match_axms false nsigma axiomMap'' cod_thms

    val axs_basis = Symtab.dest(#dom_axs_map d)
    val src = Sigmorph.src nsigma
    val trg = Sigmorph.trg nsigma
    val trg_id = Aux.name_of_thy trg

    fun check_axm_specs trg src_term nm =
     (case (Aux.assoc_string (axm_specs, nm)) of
       NONE => NONE
     | SOME tnm => (let val norm = Aux.unvarify o prop_of o Aux.beta_eta_convert
                        val (trg_term, b) = (norm (get_axiom trg tnm), true)
                                             handle e => (norm (Aux.get_thm trg tnm), false)
                    in
                      if (Alpha.term_aconv (Sign.tsig_of trg) (src_term, trg_term))
                      then SOME (tnm, b)
                      else (let val _ = Aux.warning ("Incorrect explicit axiom instantiation: " ^
                                        nm ^ " |-> " ^ tnm)
                            in
                             NONE
                            end)
                    end)) 

    fun goals sigma axmap [] = []
      | goals sigma axmap ((nm, _)::xs) =
      (case (Symtab.lookup axmap nm) of
        NONE => (let val term = Sigmorph.map_term sigma (Aux.unvarify (prop_of (get_axiom src nm)))
                     val cterm = Thm.cterm_of trg term
                 in
                  (NameSpace.base nm, cterm)::(goals sigma axmap xs)
                 end)
      | _ => goals sigma axmap xs)


    fun ins_axs_basis a trg [] = (a, trg)
      | ins_axs_basis a trg ((nm, term)::bs) =
    let val tterm = (Sigmorph.map_term nsigma term)
    in
      (case (Symtab.lookup a nm) of
              SOME i => ins_axs_basis a trg bs
            | NONE => 
            (case (check_axm_specs trg tterm nm) of
              NONE => (let val (nnm, trg') = Transfer.transfer_axiom (thy, nm) (Sigmorph.map_term nsigma)
                                             renames I trg
                           val a' = Symtab.update_new (nm, Derive_Thm.Axm nnm) a
                       in
                        ins_axs_basis a' trg' bs
                       end)
           | SOME (nnm, b) => 
              (if b then (let val a' = Symtab.update_new (nm, Derive_Thm.Axm nnm) a
                              val _ = Aux.info ("... axiom " ^ nm ^ " explicitly instantiated")
                          in
                           ins_axs_basis a' trg bs
                          end)                      
               else (let val a' = Symtab.update_new (nm, Derive_Thm.Thm nnm) a
                         val _ = Aux.info ("... axiom " ^ nm ^ " explicitly instantiated")
                     in
                      ins_axs_basis a' trg bs
                     end)))) 
     end   

in 
 if (not axiomatic) 
 then (let val gs = goals nsigma axiomMap axs_basis
       in
        case gs of
         [] => (let val ntmorph = ThyMorph { sigma = nsigma, axiomMap = axiomMap, 
                                             theoremMap = theoremMap, data = ThyMorphData td}
                in
                 ThyM (match_thms ntmorph)
                end)
       | _ => SigM (nsigma, gs)
       end)
 else (let val (axiomMap', trg') = ins_axs_basis axiomMap trg axs_basis
           val nsigma' = Sigmorph.cod_ext nsigma trg'
           val td' = get_thymorph_data nsigma'
           val ntmorph = ThyMorph { sigma = nsigma', axiomMap = axiomMap', 
                                    theoremMap = theoremMap, data = td' }
      in
       ThyM (match_thms ntmorph)
      end)
end



(* **************************** *)
(* *** Thymorph goals   ******* *)
(* **************************** *)

fun thymorph_goals syn isigma =
 let val sigma = (if syn then Sigmorph.translate_syntax Symtab.empty isigma
                  else Sigmorph.normalize Symtab.empty isigma)
     val src = Sigmorph.src sigma
     val trg = Sigmorph.trg sigma
     val (Sigmorph.SigMorphData d) = Sigmorph.data sigma
     val (ThyMorphData td) = get_thymorph_data sigma

     val axs_basis = Symtab.dest(#dom_axs_map d)
     val cod_thms = Symtab.dest(#cod_thms td)

     val axmMap = match_axms false sigma Symtab.empty cod_thms 
  
     fun mk_goals [] = []
       | mk_goals ((nm, _)::xs) =
      (case (Symtab.lookup axmMap nm) of 
        NONE => (let val term = Sigmorph.map_term sigma (Aux.unvarify (prop_of (get_axiom src nm)))
                     val cterm = Thm.cterm_of trg term
                 in
                  (NameSpace.base nm, cterm)::(mk_goals xs)
                 end)
      | _ => mk_goals xs)

     val goals = mk_goals axs_basis
 in
  (sigma, goals)
 end





(* **************************** *)
(* *** Managing thymorphs ***** *)
(* **************************** *)

val thymorphs = ref (Symtab.empty : thymorph Symtab.table)

fun add_thymorph x = (thymorphs := (Symtab.update x (!thymorphs)))
fun get_thymorph id = (Symtab.lookup (!thymorphs) id)
fun print_thymorphs () =
   let fun prnt (s, (nm, tmorph)) = let val srcnm = Aux.name_of_thy (Sigmorph.src (sigmorph tmorph))
                                        val trgnm = Aux.name_of_thy (Sigmorph.trg (sigmorph tmorph))
                                   in
                                              (s ^  (nm ^ " : " ^ srcnm ^ " --> " ^ trgnm ^ "\n"))
                                   end
   in
    Aux.foldl prnt ("", Symtab.dest(!thymorphs)) 
   end    







(* **************************** *)
(* *** Parsers          ******* *)
(* **************************** *)


(* ------------------------------------------------------------------------- *)
(* common functions for argument parsing/scanning                            *)
(* ------------------------------------------------------------------------- *)

val scan_mapsto = ((OuterParse.$$$ "|->") || (OuterParse.$$$ "\\<mapsto>"))
fun repeatd delim scan = (scan -- Scan.repeat (OuterParse.$$$ delim |-- scan ) >> op ::) || Scan.succeed []  
val scan_map_param      = (OuterParse.$$$ "(" |-- (OuterParse.string --  
                          (scan_mapsto |-- (OuterParse.string --| OuterParse.$$$ ")"))))
val scan_map_param_list = (OuterParse.$$$ "[" |-- (repeatd "," scan_map_param) 
                                               --| OuterParse.$$$ "]")

            


(* ------------------------------------------------------------------------- *)
(* set up the 'thymorph' command                                               *)
(* ------------------------------------------------------------------------- *)


val scan_col  = OuterParse.$$$ ":"
val scan_arrow = ((OuterParse.$$$ "-->") || (OuterParse.$$$ "\\<longrightarrow>"))
val scan_maps = OuterParse.$$$ "maps"
val scan_tpmp = scan_map_param_list
val scan_opmp = scan_map_param_list
val scan_thmmp = scan_map_param_list
val scan_mrphnm =OuterParse.name 
val scan_srcnm = OuterParse.name
val scan_trgnm = OuterParse.name
val scan_sigma = OuterParse.$$$ "by_sigmorph"



fun lift1 f t = let val ((tpmp, opt), t') = f t
                in
                 case opt of
                  NONE => ((SOME tpmp, NONE, NONE), t')
                | SOME (opmp, opt') => ((SOME tpmp, SOME opmp, opt'), t')
                end 

val maps_args = (lift1 ((scan_maps |-- scan_tpmp) -- 
                        (Scan.option ((scan_opmp -- (Scan.option scan_thmmp))))))




fun lift2 f t =
    let val ((((tmrphnm, srcnm), trgnm), opt_maps), t') = f t
    in 
     case opt_maps of
      NONE => ((tmrphnm, NONE, srcnm, trgnm, NONE, NONE, NONE), t')
    | SOME (tpmp, opmp, thmmp) => 
       ((tmrphnm, NONE, srcnm, trgnm, tpmp, opmp, thmmp), t')
    end

fun lift3 f t =
    let val (((tmrphnm, smrphnm), opt_thmmp), t') = f t
    in 
      ((tmrphnm, SOME smrphnm, "", "", NONE, NONE, opt_thmmp), t')
    end


val thymorph_scan_args = 
 (lift2 ((((((scan_mrphnm --| scan_col) -- scan_srcnm) --| scan_arrow) -- scan_trgnm) 
         -- (Scan.option maps_args))))
||
 (lift3 (((scan_mrphnm --| scan_sigma) -- scan_mrphnm) -- (Scan.option (scan_maps |-- scan_thmmp))))






fun thymorph_trans args =
Toplevel.theory 
( fn thy => (let val (mrphnm, opt_sigma, srcnm, trgnm, tpmp, opmp, thmmp) = args
             in
              case opt_sigma of
               NONE => (let val src = if (Aux.name_of_thy thy = srcnm) then thy else (theory srcnm)
                            val trg = if (Aux.name_of_thy thy = trgnm) then thy else (theory trgnm)
                            fun res NONE = []
                              | res (SOME x) = x
                            val tmorph = make src trg (res tpmp) (res opmp) (res thmmp)
                            val _ = Aux.info ("Theory morphism " ^ mrphnm ^ 
                                                 " : " ^ srcnm ^ " --> " ^ trgnm ^ " constructed.")
                            val _ = add_thymorph (mrphnm, tmorph)
                        in
                            thy
                        end)
             | SOME sigma_id => 
                      (let val sigma = (case (Sigmorph.get_sigmorph sigma_id) of
                                         NONE => error ("Unknown signature morphism: " ^ sigma_id)
                                        | SOME x => x)
                           val src = Sigmorph.src sigma
                           val trg = Sigmorph.trg sigma
                           val srcnm = Aux.name_of_thy src
                           val trgnm = Aux.name_of_thy trg

                           val sigma' = (if (Aux.name_of_thy thy = trgnm)
                                         then Sigmorph.cod_ext sigma thy
                                         else sigma)
                           val axMap_rep = (case thmmp of
                                             NONE => []
                                           | SOME x => x)
                           val tmorph = make_by_sigmorph sigma' axMap_rep
                           val _ = Aux.info ("Theory morphism " ^ mrphnm ^ 
                                                 " : " ^ srcnm ^ " --> " ^ trgnm ^ " constructed.")
                           val _ = add_thymorph (mrphnm, tmorph)
                       in
                            thy
                       end)
             end))

fun thymorph_parser tokens = (thymorph_scan_args tokens) |>> thymorph_trans 
                                        
val thymorph_cmd = OuterSyntax.command "thymorph" 
                   "Constructs a theory morphism between two theories by given type, operation\
                   \ and axiom maps or by given signature morphism and axiom map" 
                   OuterKeyword.thy_decl thymorph_parser;


val _ = OuterSyntax.add_keywords["by_sigmorph"];
val _ = OuterSyntax.add_parsers [thymorph_cmd];





(* ------------------------------------------------------------------------- *)
(* set up the 'translate_thm' command                                         *)
(* ------------------------------------------------------------------------- *)




val scan_tmrphnm = OuterParse.name; 
val scan_thmnm = OuterParse.string;
val scan_thmrnm = OuterParse.name;
val scan_along  = OuterParse.$$$ "along"
val scan_as  = OuterParse.$$$ "as"

val translate_thm_scan = scan_thmnm -- (scan_as |-- (scan_thmrnm -- (scan_along |-- scan_tmrphnm)))

fun translate_thm_trans args =
  Toplevel.theory (fn thy =>
  ( let val (thmnm, (thmrnm, tmrphnm)) = args 
        val tm = case (get_thymorph tmrphnm) of
                    SOME t => t
                  | NONE   => error ("Unknown theory morphism " ^ tmrphnm)
  
        val sigma = sigmorph tm
        val src   = Sigmorph.src sigma
        val srcnm = Aux.name_of_thy src
        val trg   = Sigmorph.trg sigma
        val trgnm = Aux.name_of_thy trg
        val thisnm  = Aux.name_of_thy thy

        val tm' = if (thisnm = trgnm) then (cod_ext tm Symtab.empty thy)   
                                      else tm

        val (tm'', _) = translate_thm tm' thmrnm thmnm
        val _ = add_thymorph(tmrphnm, tm'')

        val ntrg = Sigmorph.trg(sigmorph tm'')
        val newthy = (if (thisnm = Aux.name_of_thy ntrg) then ntrg
                      else thy)
    in
     newthy
    end))     


fun translate_thm_parser tokens = translate_thm_scan tokens |>> translate_thm_trans

val translate_thm_cmd = OuterSyntax.command "translate_thm" "Translate a theorem along a theory morphism" 
                         OuterKeyword.thy_decl translate_thm_parser;

val _ = OuterSyntax.add_keywords["as"]; 
val _ = OuterSyntax.add_keywords["along"];
val _ = OuterSyntax.add_parsers [translate_thm_cmd];




(* ------------------------------------------------------------------------- *)
(* set up the 't_compose' command                                         *)
(* ------------------------------------------------------------------------- *)


val scan_comp  = ((OuterParse.$$$ "o") || (OuterParse.$$$ "\\<circ>"))
val scan_tmrphnm1 = OuterParse.name; 
val scan_tmrphnm2 = OuterParse.name;
val scan_ctmrphnm = OuterParse.name;
val scan_def_eq  = ((OuterParse.$$$ "==") || (OuterParse.$$$ "\\<equiv>"))

val t_compose_scan = scan_ctmrphnm -- (scan_def_eq |-- (scan_tmrphnm2 -- (scan_comp |-- scan_tmrphnm1)))

fun t_compose_trans args =
  Toplevel.theory (fn thy =>
  ( let val (ctnm, (tnm2, tnm1)) = args 
        val t1 = case (get_thymorph tnm1) of
                      SOME t => t
                    | NONE   => error ("Unknown theory morphism " ^ tnm1)
        val t2 = case (get_thymorph tnm2) of
                      SOME t => t
                    | NONE   => error ("Unknown theory morphism " ^ tnm2)
   
        val sigma1 = sigmorph t1
        val sigma2 = sigmorph t2
        val source = Sigmorph.src sigma1
        val srcnm  = Aux.name_of_thy source
        val trg1   = Sigmorph.trg sigma1
        val trg1nm = Aux.name_of_thy trg1
        val src2   = Sigmorph.src sigma2
        val src2nm = Aux.name_of_thy src2  
        val target = Sigmorph.trg sigma2
        val trgnm  = Aux.name_of_thy target
        val thisnm  = Aux.name_of_thy thy

        val t2' = if (thisnm = trgnm) then (cod_ext t2 Symtab.empty thy)
                                      else t2
        
        val _ = Aux.info ("Composition of theory morphisms:\n" ^
                             "(" ^ tnm2 ^ " : " ^ src2nm ^ " --> " ^ trgnm ^ ") o (" 
                                 ^ tnm1 ^ " : " ^ srcnm ^ " --> " ^ trg1nm ^ ") ...")
                               
        val tc = compose t1 t2'
        val nthy   = Sigmorph.trg (sigmorph tc)
   
        val _ = Aux.info ("Theory morphism " ^ ctnm ^ " : " ^ srcnm ^ " --> " ^ trgnm ^ " constructed.") 
        val _ = add_thymorph (ctnm, tc)
    in
      if (Aux.name_of_thy nthy = thisnm) then nthy else thy
    end))


fun t_compose_parser tokens = (t_compose_scan tokens) |>> t_compose_trans
val t_compose_cmd = OuterSyntax.command "t_compose" 
                    "Define a theory morphism as a composition" 
                    OuterKeyword.thy_decl t_compose_parser;

val _ = OuterSyntax.add_parsers [t_compose_cmd];





(* ------------------------------------------------------------------------- *)
(* set up the 't_instantiate' command                                         *)
(* ------------------------------------------------------------------------- *)


val scan_mapsto = ((OuterParse.$$$ "|->") || (OuterParse.$$$ "\\<mapsto>"))
val scan_mix = Scan.option (((OuterParse.$$$ "mixfix") -- (OuterParse.$$$ ":")) |-- 
                            (OuterParse.opt_mixfix || OuterParse.opt_infix))
fun repeatd delim scan = (scan -- Scan.repeat (OuterParse.$$$ delim |-- scan )) >> op :: || Scan.succeed []; 
val scan_ren_param      = (OuterParse.$$$ "(" |-- (OuterParse.string --  
                                  (scan_mapsto |-- ((OuterParse.name -- scan_mix) --| 
                                  OuterParse.$$$ ")"))))

val scan_ren_param_list = (OuterParse.$$$ "[" |-- (repeatd "," scan_ren_param) --| OuterParse.$$$ "]");
val scan_ren_key = OuterParse.$$$ "renames" -- OuterParse.$$$ ":"
val scan_spec_types_key = OuterParse.$$$ "inst_types" -- OuterParse.$$$ ":"
val scan_spec_ops_key = OuterParse.$$$ "inst_consts" -- OuterParse.$$$ ":"
val scan_spec_axms_key = OuterParse.$$$ "inst_axioms" -- OuterParse.$$$ ":"
val scan_renames = Scan.option(scan_ren_key |-- scan_ren_param_list)
val scan_type_specs = Scan.option(scan_spec_types_key |-- scan_map_param_list)
val scan_op_specs = Scan.option(scan_spec_ops_key |-- scan_map_param_list)
val scan_axm_specs = Scan.option(scan_spec_axms_key |-- scan_map_param_list)
val scan_syn_key = Scan.option(OuterParse.$$$ "without_syntax")
val scan_ax = Scan.option(OuterParse.$$$ "axiomatic")


val instantiate_scan = OuterParse.name -- (OuterParse.$$$ "mapping" |-- (OuterParse.name -- 
                       (scan_type_specs -- (scan_op_specs -- (scan_axm_specs -- 
                       (scan_renames -- (scan_ax -- scan_syn_key)))))))

fun inst_trans args =
     Toplevel.theory (fn thy => 
 (let val (intrgnm, (mnm, (type_specs_, (op_specs_, (axm_specs_, (ren, (ax_opt, syn))))))) = args
      val thynm = Aux.name_of_thy thy

      val intrg = (if (thynm = intrgnm) then thy else (theory intrgnm))
                                     
      val tmorph = (case (get_thymorph mnm) of
                     NONE => error ("Unknown theory morphism " ^ mnm)
                   | SOME m => m)
      val src = Sigmorph.src (sigmorph tmorph) 
      val trg = Sigmorph.trg (sigmorph tmorph)
      val trgnm = Aux.name_of_thy trg
                                     
      val tmorph' = (if (trgnm = thynm) then cod_ext tmorph Symtab.empty thy 
                     else (if (trgnm mem_string (map Aux.name_of_thy (Theory.ancestors_of thy)))
                           then cod_ext tmorph Symtab.empty thy
                           else tmorph))                         
  
    val type_specs = case type_specs_ of
                        NONE => []
                      | SOME xs => xs
      val op_specs = case op_specs_ of
                        NONE => []
                      | SOME xs => xs    
      val renames = case ren of
                      NONE => Symtab.empty
                    | SOME r => Aux.foldr (uncurry Symtab.update) (r, Symtab.empty)
      val axiomatic = case ax_opt of
                        NONE => false
                      | _ => true
      val axm_specs = case axm_specs_ of
                        NONE => []
                      | SOME xs => (if ((not axiomatic) andalso (xs <> [])) 
                                    then (let val _ = Aux.warning ("Non-axiomatic: explicit axiom instantiation(s) will be ignored.")
                                          in
                                           xs
                                          end)
                                    else xs)

      val syntax = case syn of
                    NONE => true
                  | _ => false

      val _ = (if axiomatic then Aux.info ("Instantiating theory " ^ intrgnm ^ 
                                         " by theory morphism " ^ mnm ^ 
                                         " : " ^ (Aux.name_of_thy src) ^ " --> " ^ trgnm ^ " axiomatically ...")
               else  Aux.info ("Instantiating theory " ^ intrgnm ^ 
                                  " by theory morphism " ^ mnm ^ 
                                  " : " ^ (Aux.name_of_thy src) ^ " --> " ^ trgnm ^ " ..."))

      val result = instantiate axiomatic syntax type_specs op_specs axm_specs renames intrg tmorph'
      
  in
   case result of
    ThyM tmorph'' =>  (let val sigma'' = sigmorph tmorph''
                           val nsrc_nm = Aux.name_of_thy (Sigmorph.src sigma'')
                           val ntrg = Sigmorph.trg sigma''
                           val ntrg_nm = Aux.name_of_thy ntrg

                           val _ = Aux.info ("Theory morphism " ^ mnm ^ "' : " ^ 
                                                 nsrc_nm ^ " --> " ^ ntrg_nm ^ " constructed.") 
                           val _ = add_thymorph (mnm ^ "'", tmorph'')
                       in 
                         if (ntrg_nm = thynm) then ntrg else thy
                       end)
 | SigM (sigma'', goals) => (let val nsrc_nm = Aux.name_of_thy (Sigmorph.src sigma'')
                                 val ntrg = Sigmorph.trg sigma''
                                 val ntrg_nm = Aux.name_of_thy ntrg
                                
                                 fun show_goals [] = ()
                                   | show_goals ((nm, cterm)::xs) =
                                     (let val pp = (Pretty.quote o Display.pretty_cterm) cterm
                                          val _ = Output.writeln ("lemma " ^ nm ^ ":\n" ^ 
                                                                 (Pretty.string_of pp) ^ "\n")
                                      in
                                       show_goals xs
                                      end)
                                    
                                 val _ = Aux.info ("Signature morphism " ^ mnm ^ "' : " ^ 
                                                       nsrc_nm ^ " --> " ^ ntrg_nm ^ " constructed.") 
                                 val _ = Sigmorph.add_sigmorph (mnm ^ "'", sigma'')
                                 val _ = Aux.info ("To prove:")
                                 val _ = show_goals goals 
                             in 
                                if (ntrg_nm = thynm) then ntrg else thy
                             end)
  end))


fun inst_parser tokens = instantiate_scan tokens |>> inst_trans

val instantiate_cmd = OuterSyntax.command "t_instantiate" 
                      "Instantiate a theory by a theory morphism" OuterKeyword.thy_decl inst_parser;

val _ = OuterSyntax.add_keywords ["axiomatic", "without_syntax"];
val _ = OuterSyntax.add_keywords ["inst_types", "inst_consts", "inst_axioms", "renames", "mixfix", "mapping"]; 
val _ = OuterSyntax.add_parsers [instantiate_cmd];


(* *** show_thymorph_goals *** *)

fun thymorph_goals_trans (morphnm, opt_syn) =
       (Toplevel.theory 
          (fn thy => (let val thynm = Aux.name_of_thy thy
                          val morph = (case (Sigmorph.get_sigmorph morphnm) of
                                             SOME t => t
                                           | NONE => error ("Unknown signature morphism " ^ morphnm))
 
                          val target = Sigmorph.trg morph
                          val trgnm = Aux.name_of_thy target
                          val morph' = (if (thynm = trgnm) then Sigmorph.cod_ext morph thy
                                        else morph)

                          val (morph'', goals) = (case opt_syn of
                                                    NONE => thymorph_goals true morph'
                                                  | _ => thymorph_goals false morph')
                          val ntrg = Sigmorph.trg morph''

                          val nthy = (if (thynm = trgnm) then ntrg else thy)

                          fun show_goals [] = ()
                            | show_goals ((nm, cterm)::xs) =
                              (let val pp = (Pretty.quote o Display.pretty_cterm) cterm
                                   val _ = Output.writeln ("lemma " ^ nm ^ ":\n" ^ 
                                                           (Pretty.string_of pp) ^ "\n")
                               in
                                show_goals xs
                               end)

                          val _ = Sigmorph.add_sigmorph (morphnm, morph'')
                          val _ = (case goals of
                                    [] => Aux.info ("No goals.")
                                   | _ => Aux.info ("To prove:"))
                          val _ = show_goals goals
                      in
                       nthy
                      end)))

fun thymorph_goals_parser tokens = (OuterParse.name -- Scan.option(OuterParse.$$$ "without_syntax")) tokens
                                  |>> thymorph_goals_trans

val thymorph_goals_cmd = OuterSyntax.command "show_thymorph_goals" 
                     "Show goals which have to be proved in order to construct a theory morphism by the\
                     \ given signature morphism" 
                     OuterKeyword.thy_decl thymorph_goals_parser;

val _ = OuterSyntax.add_parsers [thymorph_goals_cmd];





(* *** t_print *** *)

fun t_print_trans arg =  
Toplevel.keep ( fn state =>
                      (let val res = if arg = "sigmorphs" then Sigmorph.print_sigmorphs ()
                                     else if arg = "thymorphs" then print_thymorphs ()
                                     else error ("Possible arguments are: thymorphs, sigmorphs")
                           val _ = Aux.info res
                       in
                        ()
                       end)
                    )

fun t_print_parser tokens = (OuterParse.name tokens) |>> (Toplevel.no_timing oo t_print_trans)

val t_print_cmd = OuterSyntax.improper_command "t_print" 
"Shows the list of known: sigmorphs -- signature morphisms, thymorphs -- theory morphisms" 
OuterKeyword.diag t_print_parser;

val _ = OuterSyntax.add_parsers [t_print_cmd];



end